/*
 * Creator: Naman Dixit
 * Notice: © Copyright 2020 Naman Dixit
 */

#include "nlib/nlib.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <netinet/in.h>
#include <sys/epoll.h>
#include <sys/time.h>

#define MAX_EPOLL_EVENTS 1024

# if defined(COMPILER_CLANG)
#  pragma clang diagnostic push
#   pragma clang diagnostic ignored "-Wpadded"
#   pragma clang diagnostic ignored "-Wfloat-equal"
# endif
#include "cJSON/cJSON.h"
#include "cJSON/cJSON.c"
# if defined(COMPILER_CLANG)
#  pragma clang diagnostic pop
# endif

typedef struct Command {
    Sint id;
    Sint dispatcher_socket;

    Sint memory;
} Command;

#include "time.c"
#include"socket.c"

int main(int argc, char** argv)
{
    unused_variable(argc);
    unused_variable(argv);

    Char *port = "9527";

    Sint sock_fd = socket(AF_INET, SOCK_STREAM, 0);

    struct addrinfo hints = {.ai_family = AF_INET,
                             .ai_socktype = SOCK_STREAM};
    struct addrinfo *result = NULL;
    Sint s = getaddrinfo(NULL, port, &hints, &result);
    if (s != 0) {
        fprintf(stderr, "Error: getaddrinfo: %s\n", gai_strerror(s));
        exit(-1);
    }

    // NOTE(naman): Create an epoll instance, with no flags set.
    int epoll_fd = epoll_create1(0);
    if (epoll_fd < 0) {
        perror("epoll_create1");
        exit(-1);
    }

    // NOTE(naman): Add the socket's file descriptor to the epoll set
    struct epoll_event accept_event = {.data.fd = sock_fd,
                                       .events = EPOLLIN};
    if (epoll_ctl(epoll_fd, EPOLL_CTL_ADD, sock_fd, &accept_event) < 0) {
        perror("epoll_ctl EPOLL_CTL_ADD");
        exit(-1);
    }

    // NOTE(naman): Allocate memory for epoll array
    struct epoll_event* events = calloc(MAX_EPOLL_EVENTS, sizeof(struct epoll_event));
    if (events == NULL) {
        fprintf(stderr, "Unable to allocate memory for epoll_events");
        exit(-1);
    }

    while (connect(sock_fd, result->ai_addr, result->ai_addrlen) == -1) {
        fprintf(stderr, "Error: Couldn't connect on port %s, trying again in one second...\n", port);
        sleep(1);
    }

    printf("Log: Starting communication with server on port %s...\n", port);

    U64 time_begin = timeMilli();
    U64 time_accum = 0;

    while (true) {
        // NOTE(naman): Get the fd's that are ready
        int nready = epoll_wait(epoll_fd, events, MAX_EPOLL_EVENTS, 1000);

        for (int i = 0; i < nready; i++) {
            if (events[i].events & EPOLLERR) {
                perror("epoll_wait returned EPOLLERR");
                exit(-1);
            }

            if (events[i].events & EPOLLIN) {
                B32 initialized = false;
                Size buffer_len = 0;
                Size buffer_cap = MiB(1);
                U32 buffer_expected_len = 0;
                Char *buffer = calloc(buffer_cap, sizeof(*buffer));

                Char size_bytes[4] = {0};
                Size size_bytes_count = 0;

                Command *c = NULL;

                while (true) {
                    if (initialized == false) {
                        long len = read(sock_fd,
                                        (Char*)size_bytes + size_bytes_count,
                                        4 - size_bytes_count);

                        if (len == 0) {
                            perror("read() returned zero");
                            exit(-1);
                        }

                        size_bytes_count += (Size)len;

                        if (size_bytes_count == 4) {
                            initialized = true;

                            buffer_expected_len = (U32)((size_bytes[3] <<  0U) |
                                                        (size_bytes[2] <<  8U) |
                                                        (size_bytes[1] << 16U) |
                                                        (size_bytes[0] << 24U));
                        }

                        continue;
                    } else {
                        long len = read(sock_fd,
                                        buffer + buffer_len,
                                        buffer_expected_len - buffer_len);

                        buffer_len += (Size)len;

                        if (buffer_expected_len == buffer_len) {
                            cJSON_Print(cJSON_Parse(buffer));
                            printf("%.*s", (int)buffer_len, buffer);

                            const Char *json_error = NULL;
                            cJSON *root = cJSON_ParseWithOpts(buffer, &json_error, true);

                            if (root == NULL) {
                                // TODO(naman): Error
                            } else {
                                c = calloc(1, sizeof(*c));

                                c->id = cJSON_GetObjectItem(root, "id")->valueint;
                                c->dispatcher_socket = cJSON_GetObjectItem(root, "dispatcher_socket")->valueint;

                                c->memory = cJSON_GetObjectItem(root, "memory")->valueint;
                            }

                            free(buffer);
                            break;
                        }
                    }
                }

                int memory = 0;

                FILE *meminfo = fopen("/proc/meminfo", "r");
                Char line[256] = {0};
                while(fgets(line, sizeof(line), meminfo)) {
                    if (sscanf(line, "MemTotal: %d kB", &memory) == 1) {
                        fclose(meminfo);
                        break;
                    }
                }

                Char *output = NULL;

                sbufPrint(output, "    ");
                sbufPrint(output, "{\n\"id\": %d", c->id);
                sbufPrint(output, ",\n\"dispatcher_socket\": %d", c->dispatcher_socket);
                sbufPrint(output, ",\n\"type\": \"response\"");

                if (memory >= c->memory) {
                    sbufPrint(output, ",\n\"success\": %d\n", 1);
                } else {
                    sbufPrint(output, ",\n\"success\": %d\n", 0);
                }

                sbufPrint(output, "\n}\n");
                Size output_len = strlen(output);

#if defined(ENDIAN_LITTLE)
                U32 json_len = (U32)output_len - 4;
                U32 json_len_be = swap_endian(json_len);
                output[0] = ((Char*)&json_len_be)[0];
                output[1] = ((Char*)&json_len_be)[1];
                output[2] = ((Char*)&json_len_be)[2];
                output[3] = ((Char*)&json_len_be)[3];
#endif

                socketWrite(output, output_len, sock_fd);
            }
        }

        { // Send a heartbeat message if it is time to do so
            U64 time_new = timeMilli();
            U64 time_passed = time_new - time_begin;
            time_begin = time_new;
            time_accum += time_passed;

            if (time_accum >= 1000) {
                time_accum = 0;

                int memory = 0;

                FILE *meminfo = fopen("/proc/meminfo", "r");
                Char line[256] = {0};
                while(fgets(line, sizeof(line), meminfo)) {
                    if (sscanf(line, "MemTotal: %d kB", &memory) == 1) {
                        fclose(meminfo);
                        break;
                    }
                }

                Char *output = NULL;

                sbufPrint(output, "    ");
                sbufPrint(output, "{");
                sbufPrint(output, " \n\"type\": \"heartbeat\"");
                sbufPrint(output, ",\n\"memory\": %d", memory);

                sbufPrint(output, "\n}\n");
                Size output_len = strlen(output);

#if defined(ENDIAN_LITTLE)
                U32 json_len = (U32)output_len - 4;
                U32 json_len_be = swap_endian(json_len);
                output[0] = ((Char*)&json_len_be)[0];
                output[1] = ((Char*)&json_len_be)[1];
                output[2] = ((Char*)&json_len_be)[2];
                output[3] = ((Char*)&json_len_be)[3];
#endif

                socketWrite(output, output_len, sock_fd);
            }
        }
    }
}
